#include "include_2col.h"

/**********************************************************************/
double dtime()
{
  // timing

    double tseconds = 0.0;
    struct timeval mytime;
    gettimeofday(&mytime,(struct timezone*)0);
    tseconds = (double)(mytime.tv_sec + mytime.tv_usec*1.0e-6);
    return( tseconds );
}


/**********************************************************************/
int main(int argc, char *argv[] ) {

  //* multiply 3-component complex vectors by 3x3 complex matricies



  su3vec vecA;
  su3vec vecB;
  su3mat matM;


  int i,j;
  int rep;
  int num_reps=400;

  int indx;
  int tindex;

  int hi, lo, offset;
  double tstart, tstop, ttime;
  double gflops = 0.0;
  double gflop_per_sec=0.0;
  double gbytes;
  double actual_gbytes;
  double bandwidth=0.0;
  double intensity;
  float block_data_size;
  float total_data_size;

  int numthreads;
  
  int ops_per_iter = 8;
  int bytes_per_iter = 32;
  
  int blocksize;

   //
    // initialize the compute arrays 
    //
    //

    // Start some threads to get the number of threads from 
    //OMP_NUM_THREADS environment variable
#pragma omp parallel
#pragma omp master
  numthreads = omp_get_num_threads();


  // data size in MB
  total_data_size=(2*sizeof(su3vec)+sizeof(su3mat))/(1024*1024.);
  printf("\n\ntotal data size = %g MB\n", total_data_size);



  //determine block size
  blocksize=VOLUME/numthreads;
  
  if(numthreads*blocksize<VOLUME){
    // do we have to re-size to get the extra??

    printf("resizing blocks %d -> ",blocksize);
    blocksize++;
    printf("%d\n",blocksize);
  }
  printf("\n\n\nBLOCKSIZE\t%d\n",blocksize);
  printf("block data size\t%g MB\n",(float)blocksize*total_data_size/VOLUME);

  


  // INITIALIZATION
  for(i=0; i< SU3VEC_SIZE; i++){
    // threading to speed up initialization
    // not part of the test
#pragma omp parallel for 
    for(j=0; j< VOLUME; j++){

      // initialize vecA with some stuff
      vecA.vec[i].re[j]=(float)(i+2);
      vecA.vec[i].im[j]=(float)(i-2);

      // initialize vecA with zeros
      vecB.vec[i].re[j]=0.0;
      vecB.vec[i].im[j]=0.0;
    }// volume loop
  }// i suvec loop
  

  // initialize with some sample SU(3) fields
  init_su3mats(&matM);


  tstart = dtime();

  // do some repeats so we have enought to time 
  for(rep=0; rep < num_reps; rep++){


 
#pragma omp parallel for private(indx, offset, lo, hi, i, j)
    for(tindex=0; tindex<numthreads; tindex++){
      offset=tindex*blocksize;
      
      lo=offset;
      hi=offset+blocksize;
      if (hi > VOLUME){
	hi = VOLUME;
      }

      //TODO: comment this line to block  the loop
      //      for(indx=lo; indx<hi ; indx++){	  

	for(i=0; i<SU3VEC_SIZE; i++){
	  for(j=0; j<SU3VEC_SIZE; j++){
	    
	  
	   //TODO: uncomment this line to block	  
	    for(indx=lo; indx<hi ; indx++){
	  


	   //#ops is 4 per iteration
	   vecB.vec[i].re[indx] =  vecB.vec[i].re[indx] +
	     matM.mat[3*i+j].re[indx] * vecA.vec[j].re[indx];

	   vecB.vec[i].re[indx] = vecB.vec[i].re[indx] -
             matM.mat[3*i+j].im[indx] * vecA.vec[j].im[indx];


	   vecB.vec[i].im[indx] = vecB.vec[i].im[indx] +
             matM.mat[3*i+j].re[indx] * vecA.vec[j].im[indx];
           vecB.vec[i].im[indx] = vecB.vec[i].im[indx] +
             matM.mat[3*i+j].im[indx] * vecA.vec[j].re[indx];

	   //TODO: uncomment this to move the closing brace (optional)
	   /*
	}// indx
	   */
	}// j-loop
      }  // i-loop

      // Comment this line to move the closing brace (optional)
      }// indx loop
    }//tindex

  }//reps

    tstop = dtime();
  //elasped time
  ttime = tstop - tstart;
  //
  

  // # of gigaflops we just calculated  
  gflops = (double)( 1.0e-9*num_reps*VOLUME*ops_per_iter*9);
  gbytes = (double)( 1.0e-9*num_reps*VOLUME*bytes_per_iter*9);

  gflop_per_sec= gflops/ttime ;
  bandwidth = gbytes/ttime ;

  printf(" threads = %d\n",numthreads);
  printf("total gflops: %f\n", gflops);
  printf("total gbytes: %f\n", gbytes);
  printf ("elapsed time (s) = %f\n",ttime);
  printf("PERF:\tGflop/sec = %f\n",gflop_per_sec);
  printf("Bandwidth GB/sec = %f\n",bandwidth);


  printf("op/iter = %d\n",ops_per_iter);
  printf("byte/iter = %d\n",bytes_per_iter);


  printf("Intensity (flops/byte)= %f\n", (float)ops_per_iter/((float)bytes_per_iter));

}
